package com.streambase.liveview.sample.restaction;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.zip.GZIPInputStream;
import java.util.zip.Inflater;
import java.util.zip.InflaterInputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.streambase.liveview.client.AbstractTupleEvent;
import com.streambase.liveview.client.LiveViewException;
import com.streambase.liveview.client.alert.AlertRule;
import com.streambase.liveview.client.alert.ExecuteJavaAction;
import com.streambase.liveview.client.alert.actions.ExecJavaAction;
import com.streambase.sb.util.Util;

/*
 * This class implements the LDM ExecJavaAction action and will do an HTTP request with 
 * the supplied parameters.
 */
public class RESTAction implements ExecJavaAction {

	private Logger logger = LoggerFactory.getLogger(this.getClass().getName());
	
    public static enum HTTPRequestType {
    	HTTP_REQUEST_TYPE_GET("GET"),
        HTTP_REQUEST_TYPE_HEAD("HEAD"),
        HTTP_REQUEST_TYPE_POST("POST"),
        HTTP_REQUEST_TYPE_PUT("PUT"),
        HTTP_REQUEST_TYPE_PATCH("PATCH"),
        HTTP_REQUEST_TYPE_DELETE("DELETE");

        private final String request;

        public static HTTPRequestType FromString(String s) throws IllegalArgumentException {
        	switch (s) {
        	case "GET":
        		return HTTP_REQUEST_TYPE_GET;
        	case "HEAD":
        		return HTTP_REQUEST_TYPE_HEAD;
        	case "POST":
        		return HTTP_REQUEST_TYPE_POST;
        	case "PUT":
        		return HTTP_REQUEST_TYPE_PUT;
        	case "DELETE":
        		return HTTP_REQUEST_TYPE_DELETE;
        	case "PATCH":
        		return HTTP_REQUEST_TYPE_PATCH;
        	default:
        		throw new IllegalArgumentException(String.format("Unknown HTTP request type: %s", s));
        	}
        }
        
        private HTTPRequestType(String s) {
            request = s;
        }

        @Override
        public String toString() {
            return request;
        }
    }
    
    /* 
     * These are the parameters the Action can set
     */
    private final String HTTPRequestTypeOption_name="RequestType";
    private final String basicAuthUsername_name="username";
    private final String basicAuthPassword_name="password";
    private final String urlString_name="uri";
    private final String urlParams_regxname="urlParam.*";
    private final String HTTPHeaders_regxname="HTTPHeader.*";
    private final String verbose_name="verbose";
    
    private HTTPRequestType HTTPRequestTypeOption=null;
    private String basicAuthUsername=null;
    private String basicAuthPassword=null;
    private String urlString=Util.getSystemProperty("liveview.sample.action.url");
    private String urlParams=null;
    private boolean verbose=true;
    
    private boolean enableBasicAuth=false;
    
	@Override
	public boolean action(AlertRule alertRule, ExecuteJavaAction action, AbstractTupleEvent event) throws LiveViewException {	
		Map<String, String> param=action.getParameters();
		HashMap<String, String> httpHeaders = new HashMap<String, String>();
		
		StringBuilder params = new StringBuilder();
		for (String k : param.keySet()) {
			if (urlString_name.equals(k)) {
				urlString=param.get(k);
				continue;
			}
			if (basicAuthUsername_name.equals(k)) {
				enableBasicAuth=true;
				basicAuthUsername=param.get(k);
				continue;
			}
			if (basicAuthPassword_name.equals(k)) {
				basicAuthPassword=param.get(k);
				continue;
			}
			if (HTTPRequestTypeOption_name.equals(k)) {
				HTTPRequestTypeOption= HTTPRequestType.FromString(param.get(k));
				continue;
			}
			if (verbose_name.equals(k)) {
				verbose=new Boolean(param.get(k));
				continue;
			}
			if (k.matches(urlParams_regxname)) {
				params.append(param.get(k));
				continue;
			}
			if (k.matches(HTTPHeaders_regxname)) {
				String [] header=param.get(k).split(":");
				httpHeaders.put(header[0], header[1]);
				continue;
			}
		}
		
		if (urlString==null || HTTPRequestTypeOption==null) {
			logger.warn(String.format("ExecJava Alert rule: %s must set at least %s and %s", alertRule.getId(), urlString_name, HTTPRequestTypeOption_name));
			return false;
		}
		
		urlParams = params.toString();
		
		return doURL(urlString, urlParams, httpHeaders, 0, 0);
	}

   private boolean doURL(String urlString, String urlParams, HashMap<String, String> httpHeaders, int useConnectTimeout, int useReadTimeout) {
        URL url;
        HttpURLConnection connection = null;
        try {
            BufferedReader reader;
            StringBuffer result = new StringBuffer();
            url = new URL(urlString);
            connection = (HttpURLConnection) url.openConnection();
            connection.setConnectTimeout(useConnectTimeout);
            connection.setReadTimeout(useReadTimeout);
            connection.setRequestMethod(HTTPRequestTypeOption.toString());
            if (httpHeaders != null) {
                for (Entry<String, String> headers : httpHeaders.entrySet()) {
                    if (headers.getKey() != null && headers.getKey().length() > 0 && headers.getValue() != null && headers.getValue().length() > 0) {
                        connection.setRequestProperty(headers.getKey(), headers.getValue());
                    }
                }
            }
            connection.setRequestProperty("Accept-Encoding", "gzip, deflate");
            setAuthHeader(connection);
            if ((HTTPRequestTypeOption.equals(HTTPRequestType.HTTP_REQUEST_TYPE_POST) || HTTPRequestTypeOption.equals(HTTPRequestType.HTTP_REQUEST_TYPE_PUT)) && urlParams != null && urlParams.length() > 0) {
                connection.setDoInput(true);
                connection.setDoOutput(true);
                OutputStream os = connection.getOutputStream();
                BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(os));
                writer.write(urlParams);
                writer.close();
            }

            int responseCode = connection.getResponseCode();
            StringBuilder response= new StringBuilder();
            if (responseCode >= 200 && responseCode < 300) {                
                InputStream resultingInputStream = createEncodingInputStream(connection, connection.getInputStream());
                
                Iterator<Entry<String, List<String>>> headers = connection.getHeaderFields().entrySet().iterator();
                while (headers.hasNext()) {
                    Entry<String, List<String>> entry = headers.next();
                    if (entry != null && entry.getKey() != null && entry.getKey().length() > 0) {
                        response.append(entry.getKey());
                        response.append(" ");
                        response.append(entry.getValue());
                        response.append("\n");
                    }
                }
                if (verbose) {
	                // Dump the headers 
	                logger.info(response.toString());
                }
                reader = new BufferedReader(new InputStreamReader(resultingInputStream));
                readBufferedReader(result, reader);
                reader.close();
                if (verbose) {
	                // Dump the result
	                logger.info(result.toString());
                }
                return true;
            } else {
                if (connection.getErrorStream() != null) {
                    reader = new BufferedReader(new InputStreamReader(createEncodingInputStream(connection, connection.getErrorStream())));
                    readBufferedReader(result, reader);
                    reader.close();
                } else {
                    result.append(connection.getResponseMessage() == null ? "" : connection.getResponseMessage());
                }
                String message = String.format("For URL: %s HTTP [" + responseCode + "] Error: %s", urlString, result.toString());
                logger.warn(message);
                return false;
            }            
        } catch (Exception e) {
            String message = String.format("Exception reading HTTP URL %s: %s", urlString, e.getMessage());
            logger.warn(message, logger.isDebugEnabled() ? e : null);
            return false;
        } finally {
            if (connection != null) {
                connection.disconnect();
            }
        }
    }

    private void setAuthHeader(HttpURLConnection connection) {
        if (enableBasicAuth) {
            connection.setRequestProperty("Authorization", "Basic " + com.streambase.sb.util.Base64.encodeBytes((basicAuthUsername + ":" + basicAuthPassword).getBytes()));
        }
    }

    private InputStream createEncodingInputStream(HttpURLConnection connection, InputStream inputStream) throws IOException {
        String encoding = connection.getContentEncoding();
        // create the appropriate stream wrapper based on the encoding type
        if (encoding != null && encoding.equalsIgnoreCase("gzip")) {
        	return new GZIPInputStream(inputStream);
        } else if (encoding != null && encoding.equalsIgnoreCase("deflate")) {
            return new InflaterInputStream(inputStream, new Inflater(true));
        }
        return inputStream;        
    }
    
    private void readBufferedReader(StringBuffer result, BufferedReader reader) throws IOException {
        int character;
        while ((character = reader.read()) != -1) {
            result.append((char)character);
        }
    }
}
